##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import datetime
import os
import subprocess
import sys
import time
import glob
import SlapOSControler
import SlapOSMasterCommunicator
import json
import time
import shutil
import logging
import string
import random
import Utils
from ProcessManager import SubprocessError, ProcessManager, CancellationError
from subprocess import CalledProcessError
from Updater import Updater
from erp5.util import taskdistribution
# for dummy slapos answer
import signal

# max time to instance changing state: 2 hour
MAX_INSTANCE_TIME = 60*60*2
# max time to register instance to slapOSMaster: 5 minutes
MAX_CREATION_INSTANCE_TIME = 60*10
# max time for a test: 1 hour
MAX_TEST_CASE_TIME = 60*60

class ScalabilityTestRunner():
  def __init__(self, testnode):
    self.testnode =  testnode
    self.log = self.testnode.log
    
    self.slapos_controler = SlapOSControler.SlapOSControler(
                                  self.testnode.working_directory,
                                  self.testnode.config,
                                  self.log)
    # Create the slapos account configuration file and dir
    key = self.testnode.test_suite_portal.getSlaposAccountKey()
    certificate = self.testnode.test_suite_portal.getSlaposAccountCertificate()

    # Get Slapos Master Url
    self.slapos_url = ''
    try:
      self.slapos_url = self.testnode.test_suite_portal.getSlaposUrl()
      if not self.slapos_url:
        self.slapos_url = self.testnode.config['server_url']
    except:
      self.slapos_url = self.testnode.config['server_url']
    
    # Get Slapos Master url used for api rest (using hateoas)
    self.slapos_api_rest_url = self.testnode.test_suite_portal.getSlaposHateoasUrl()

    self.log("SlapOS Master url is: %s" %self.slapos_url)
    self.log("SlapOS Master hateoas url is: %s" %self.slapos_api_rest_url)
    
    self.key_path, self.cert_path, config_path = self.slapos_controler.createSlaposConfigurationFileAccount(
                                        key, certificate, self.slapos_url, self.testnode.config)
    self.slapos_communicator = None
    # Dict containing used to store which SR is not yet correctly installed.
    # looks like: {'comp_id1':'SR_urlA', 'comp_id2':'SR_urlA',..}
    self.remaining_software_installation_dict = {}
    
    # Protection to prevent installation of softwares after checking
    self.authorize_supply = True
    self.authorize_request = False
    # Used to simulate SlapOS answer (used as a queue)
    self.last_slapos_answer = []
    self.last_slapos_answer_request = []
    
  def _prepareSlapOS(self, software_path, computer_guid, create_partition=0):
    # create_partition is kept for compatibility
    """
    A proxy to supply : Install a software on a specific node
    """
    self.log("testnode, supply : %s %s", software_path, computer_guid)
    if self.authorize_supply :
      self.remaining_software_installation_dict[computer_guid] = software_path
      self.slapos_controler.supply(software_path, computer_guid)
      # Here make a request via slapos controler ?
      return {'status_code' : 0}                                          
    else:
      raise ValueError("Too late to supply now. ('self.authorize_supply' is False)")
      return {'status_code' : 1}     

  def _generateInstanceTitle(self, test_suite_title):
    """
    Generate an instance title using various parameter
    TODO : add some verification (to don't use unexisting variables)
    """
    instance_title = "Scalability-"
    instance_title += "("+test_suite_title+")-"
    instance_title += str(self.involved_nodes_computer_guid).replace("'","")
    instance_title += "-"+str(datetime.datetime.now().isoformat())+"-"
    instance_title += "timestamp="+str(time.time())
    return instance_title

  def _generateInstanceXML(self, software_configuration,
                      test_result, test_suite):
    """
    Generate a complete scalability instance XML configuration
    """
    config_cluster = software_configuration.copy()
    config = {'cluster':config_cluster}
    config.update({'scalability-launcher-computer-guid':self.launcher_nodes_computer_guid[0]})
    config.update({'scalability-launcher-title':'MyTestNodeTitle'})
    config.update({'test-result-path':test_result.test_result_path})
    config.update({'test-suite-revision':test_result.revision})
    config.update({'test-suite':test_suite})
    config.update({'test-suite-master-url':self.testnode.config['test_suite_master_url']})
    return config
  
  def _createInstance(self, software_path, software_configuration, instance_title,
                      test_result, test_suite):
    """
    Create scalability instance
    """
    if self.authorize_request:
      config = self._generateInstanceXML(software_configuration,
                                    test_result, test_suite)
      self.log("testnode, request : %s", instance_title)
      config = json.dumps(config)
      self.slapos_controler.request(instance_title, software_path,
                             "test", {"_" : config},
                             self.launcher_nodes_computer_guid[0])
      self.authorize_request = False
      return {'status_code' : 0}                                          
    else:
      raise ValueError("Softwares release not ready yet to launch instan\
ces or already launched.")
      return {'status_code' : 1}  
      
  def prepareSlapOSForTestNode(self, test_node_slapos=None):
    """
    We will build slapos software needed by the testnode itself,
    """
    if self.testnode.test_suite_portal.isMasterTestnode(
                           self.testnode.config['test_node_title']):
      pass
    return {'status_code' : 0} 

  # Dummy slapos answering
  def _getSignal(self, signal, frame):
    self.log("Dummy SlapOS Master answer received.")
    self.last_slapos_answer.append(True)
  def _prepareDummySlapOSAnswer(self):
    self.log("Dummy slapOS answer enabled, send signal to %s (kill -10 %s) to simu\
late a SlapOS (positive) answer." %(str(os.getpid()),str(os.getpid()),))
    signal.signal(signal.SIGUSR1, self._getSignal)
  def _comeBackFromDummySlapOS(self):
    self.log("Dummy slapOS answer disabled, please don't send more signals.")
    # use SIG_USR (kill)
    signal.signal(signal.SIGUSR1, signal.SIG_DFL)
  def simulateSlapOSAnswer(self):
    if len(self.last_slapos_answer)==0:
      return False
    else:
      return self.last_slapos_answer.pop()
  # /Dummy slapos answering
    
  def isSoftwareReleaseReady(self, software_url, computer_guid):
    """
    Return true if the specified software on the specified node is installed.
    This method should communicates with SlapOS Master.
    """
    # TODO : implement -> communication with SlapOS master
    # this simulate a SlapOS answer
    return self.simulateSlapOSAnswer()
  
  def remainSoftwareToInstall(self):
    """
    Return True if it remains softwares to install, otherwise return False
    """
    # Remove from grid installed software entries
    for computer_guid, software_path in self.remaining_software_installation_dict.items():
      if self.isSoftwareReleaseReady(software_path, computer_guid):
        del self.remaining_software_installation_dict[computer_guid]
    # Not empty grid means that all softwares are not installed
    return len(self.remaining_software_installation_dict) > 0

  def _updateInstanceXML(self, software_configuration, instance_title,
                      test_result, test_suite):
    """
    Just a proxy to SlapOSControler.updateInstanceXML.
    """
    config = self._generateInstanceXML(software_configuration,
                                  test_result, test_suite)
    config = json.dumps(config)
    self.log("testnode, updateInstanceXML : %s", instance_title)
    self.slapos_controler.updateInstanceXML(instance_title, {"_" : config})
    return {'status_code' : 0} 

  def _waitInstance(self, instance_title, state, max_time=MAX_INSTANCE_TIME):
    """
    Wait for 'max_time' an instance specific state
    """
    self.log("Wait for instance state: %s" %state)
    start_time = time.time()
    while (not self.slapos_communicator.isHostingSubscriptionReady(instance_title, state)
         and (max_time > (time.time()-start_time))):
      self.log("Instance(s) not in %s state yet." % state)
      time.sleep(15)
    if (time.time()-start_time) > max_time:
      error_message = "Instance '%s' not '%s' after %s seconds" %(instance_title, state, str(time.time()-start_time))
      self.log(error_message)
      self.log("Do you use instance state propagation in your project?")
      self.log("Instance '%s' will be stopped and test avorted." %instance_title)
      # What if we wanted to stop ?
      self.slapos_controler.stopInstance(instance_title)
      # XXX: _waitInstance call here ? recursive call ?
      # XXX: sleep 60 seconds.
      time.sleep(60) 
      raise ValueError(error_message)
    self.log("Instance correctly '%s' after %s seconds." %(state, str(time.time()-start_time)))

  def _waitInstanceCreation(self, instance_title, max_time=MAX_CREATION_INSTANCE_TIME):
    """
    Wait for 'max_time' the instance creation
    """
    self.log("Wait for instance creation")
    start_time = time.time()
    while ( not self.slapos_communicator.isRegisteredHostingSubscription(instance_title) \
         and (max_time > (time.time()-start_time)) ):
      time.sleep(5)
    if (time.time()-start_time) > max_time:
      raise ValueError("Instance '%s' not found after %s seconds" %(instance_title, max_time))
    self.log("Instance found on slapOSMaster")

  def prepareSlapOSForTestSuite(self, node_test_suite):
    """
    Install testsuite softwares
    """
    self.log('prepareSlapOSForTestSuite')
    # Define how many time this method can take
    max_time = 3600*10*1.0 # 10 hours
    interval_time = 60
    start_time = time.time()
    # Create a communicator with slapos
    self.log("creating SlapOs Master communicator...")
    self.slapos_communicator = SlapOSMasterCommunicator.SlapOSMasterCommunicator(
                                  self.cert_path,
                                  self.key_path,
                                  self.log,
                                  self.slapos_api_rest_url)
    # Only master testnode must order software installation
    if self.testnode.test_suite_portal.isMasterTestnode(
            self.testnode.config['test_node_title']):
      # Get from ERP5 Master the configuration of the cluster for the test
      test_configuration = Utils.deunicodeData(
          json.loads(self.testnode.test_suite_portal.generateConfiguration(
                      node_test_suite.test_suite_title)
                    )
        )
      self.involved_nodes_computer_guid = test_configuration['involved_nodes_computer_guid']
      self.launchable = test_configuration['launchable']
      self.error_message = test_configuration['error_message']
      self.randomized_path = test_configuration['randomized_path']
      # Avoid the test if it is not launchable
      if not self.launchable:
        self.log("Test suite %s is not actually launchable with \
  the current cluster configuration." %(node_test_suite.test_suite_title,))
        self.log("ERP5 Master indicates : %s" %(self.error_message,))
        # error : wich code to return ?
        return {'status_code' : 1}

      involved_nodes_computer_guid = test_configuration['involved_nodes_computer_guid']
      configuration_list = test_configuration['configuration_list']
      node_test_suite.edit(configuration_list=configuration_list)
      self.launcher_nodes_computer_guid = test_configuration['launcher_nodes_computer_guid']
      
      # Create an obfuscated link to the testsuite directory
      path_to_suite = os.path.join(
                      self.testnode.config['working_directory'],
                      node_test_suite.reference)
      self.obfuscated_link_path = os.path.join(
                      self.testnode.config['software_directory'],
                      self.randomized_path)
      if ( not os.path.lexists(self.obfuscated_link_path) and
           not os.path.exists(self.obfuscated_link_path) ) :
        try :
          os.symlink(path_to_suite, self.obfuscated_link_path)
          self.log("testnode, Symbolic link (%s->%s) created."
                   %(self.obfuscated_link_path, path_to_suite))
        except :
          self.log("testnode, Unable to create symbolic link to the testsuite.")
          raise ValueError("testnode, Unable to create symbolic link to the testsuite.")
      self.log("Sym link : %s %s" %(path_to_suite, self.obfuscated_link_path))
      
      # Construct the ipv6 obfuscated url of the software profile reachable from outside
      self.reachable_address = os.path.join(
        "https://","["+self.testnode.config['httpd_ip']+"]"+":"+self.testnode.config['httpd_software_access_port'],
        self.randomized_path)
      self.reachable_profile = os.path.join(self.reachable_address, "software.cfg")

      # Write the reachable address in the software.cfg file,
      # by replacing <obfuscated_url> occurences by the current reachable address.
      software_file = open(node_test_suite.custom_profile_path, "r")
      file_content = software_file.readlines()
      new_file_content = []
      for line in file_content:
        new_file_content.append(line.replace('<obfuscated_url>', self.reachable_address))
      software_file.close()
      os.remove(node_test_suite.custom_profile_path)
      software_file = open(node_test_suite.custom_profile_path, "w")
      for line in new_file_content:
        software_file.write(line)
      software_file.close()
      self.log("Software reachable profile path is : %s "
                              %(self.reachable_profile,))

      # Ask for SR installation
      for computer_guid in self.involved_nodes_computer_guid:
        self._prepareSlapOS(self.reachable_profile, computer_guid) 
      # From the line below we would not supply any more softwares
      self.authorize_supply = False
      # TODO : remove the line below wich simulate an answer from slapos master
      self._prepareDummySlapOSAnswer()
      # Waiting until all softwares are installed
      while ( self.remainSoftwareToInstall() 
         and (max_time > (time.time()-start_time))):
        self.log("Master testnode is waiting\
 for the end of all software installation (for %ss) PID=%s.",
          str(int(time.time()-start_time)), str(os.getpid()))
        time.sleep(interval_time)
      # TODO : remove the line below wich simulate an answer from slapos master
      self._comeBackFromDummySlapOS()
      if self.remainSoftwareToInstall() :
        # All softwares are not installed, however maxtime is elapsed, that's a failure.
        return {'status_code' : 1}
      self.authorize_request = True
      self.log("Softwares installed")
      # Launch instance
      self.instance_title = self._generateInstanceTitle(node_test_suite.test_suite_title)
      try:
        self._createInstance(self.reachable_profile, configuration_list[0],
                              self.instance_title, node_test_suite.test_result, node_test_suite.test_suite)
        self.log("Scalability instance requested.")
      except:
        self.log("Unable to launch instance")
        raise ValueError("Unable to launch instance")
      self.log("Waiting for instance creation..")
      self._waitInstanceCreation(self.instance_title)
      return {'status_code' : 0}
    return {'status_code' : 1}

  def runTestSuite(self, node_test_suite, portal_url):
    if not self.launchable:
      self.log("Current test_suite is not actually launchable.")
      return {'status_code' : 1} # Unable to continue due to not realizable configuration
    configuration_list = node_test_suite.configuration_list
    test_list = range(0, len(configuration_list))
    # create test_result
    test_result_proxy = self.testnode.portal.createTestResult(
      node_test_suite.revision, test_list,
      self.testnode.config['test_node_title'],
      True, node_test_suite.test_suite_title,
      node_test_suite.project_title)
  
    count = 0
    error_message = None

    # Each cluster configuration are tested
    for configuration in configuration_list:

      # First configuration doesn't need XML configuration update.
      if count > 0:
        # Stop instance
        self.slapos_controler.stopInstance(self.instance_title)
        self._waitInstance(self.instance_title, 'stopped')
        # Update instance XML configuration 
        self._updateInstanceXML(configuration, self.instance_title,
                      node_test_suite.test_result, node_test_suite.test_suite)
        self._waitInstance(self.instance_title, 'started')
        # Start instance
        self.slapos_controler.startInstance(self.instance_title)
        
      # XXX: Dirty hack used to force haproxy to restart in time
      # with all zope informations.
      self._waitInstance(self.instance_title, 'started')
      self.slapos_controler.stopInstance(self.instance_title)
      self._waitInstance(self.instance_title, 'stopped')
      self.slapos_controler.startInstance(self.instance_title)
      ##########################################################
        
      self._waitInstance(self.instance_title, 'started')
      
      # Start only the current test
      exclude_list=[x for x in test_list if x!=test_list[count]]
      count += 1
      test_result_line_proxy = test_result_proxy.start(exclude_list)

      # 
      if test_result_line_proxy == None :
        error_message = "Test case already tested."
        break

      self.log("Test for count : %d is in a running state." %count)

      # Wait for test case ending
      test_case_start_time = time.time()
      while test_result_line_proxy.isTestCaseAlive() and \
            test_result_proxy.isAlive() and \
            time.time() - test_case_start_time < MAX_TEST_CASE_TIME:
        time.sleep(15)

      # Max time limit reach for current test case: failure.
      if test_result_line_proxy.isTestCaseAlive():
        error_message = "Test case during for %s seconds, too long. (max: %s seconds). Test failure." \
                            %(str(time.time() - test_case_start_time), MAX_TEST_CASE_TIME)
        break

      # Test cancelled, finished or in an undeterminate state.
      if not test_result_proxy.isAlive():
        # Test finished
        if count == len(configuration_list):
          break
        # Cancelled or in an undeterminate state.
        error_message = "Test cancelled or undeterminate state."
        break

    # Stop current instance
    self.slapos_controler.stopInstance(self.instance_title)
    self._waitInstance(self.instance_title, 'stopped')

    # Delete old instances
    self._cleanUpOldInstance()

    # If error appears then that's a test failure.    
    if error_message:
      test_result_line_proxy.stop(error_count=1, failure_count=1,
                                  stdout=error_message, stderr=error_message)
      test_result_proxy.reportFailure(stdout=error_message)
      self.log("Test Failed.")
      return {'status_code' : 1, 'error_message': error_message} 
    # Test is finished.
    self.log("Test finished.")
    return {'status_code' : 0}

  def _cleanUpOldInstance(self):
    self.log("_cleanUpOldInstance")

    # Get title and link list of all instances
    instance_dict = self.slapos_communicator.getHostingSubscriptionDict()
    instance_to_delete_list = []
    outdated_date = datetime.datetime.fromtimestamp(time.time()) - datetime.timedelta(days=2)

    # Select instances to delete
    for title,link in instance_dict.items():
      # Instances created by testnode contains "Scalability-" and
      # "timestamp=" in the title.
      if "Scalability-" in title and "timestamp=" in title:
        # Get timestamp of the instance creation date
        foo, timestamp = title.split("timestamp=")
        creation_date = datetime.datetime.fromtimestamp(float(timestamp))
        # Test if instance is older than the limit
        if creation_date < outdated_date:
          instance_to_delete_list.append((title,link))
    
    for title,link in instance_to_delete_list:
      # Get instance information
      instance_information_dict = self.slapos_communicator.getHostingSubscriptionInformationDict(title)
      # Delete instance
      if instance_information_dict:
        if instance_information_dict['status'] != 'destroyed':
          self.slapos_controler.request(
              instance_information_dict['title'],
              instance_information_dict['software_url'],
              software_type=instance_information_dict['software_type'],
              computer_guid=instance_information_dict['computer_guid'],
              state='destroyed'
          )
          self.log("Instance '%s' deleted." %instance_information_dict['title'])

  def _cleanUpNodesInformation(self):
    self.involved_nodes_computer_guid = []
    self.launcher_nodes_computer_guid = []
    self.remaining_software_installation_dict = {}
    self.authorize_supply = True
    self.authorize_request = False

  def getRelativePathUsage(self):
    """
    Used by the method testnode.constructProfile() to know
    if the software.cfg have to use relative path or not.
    """
    return True



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import ConfigParser
import argparse
import logging
import logging.handlers
import os
import pkg_resources

from testnode import TestNode

CONFIG = {
  'computer_id': 'COMPUTER',
  'partition_reference': 'test0',
}

def main(*args):
  parser = argparse.ArgumentParser()
  parser.add_argument("configuration_file", nargs=1, type=argparse.FileType(),
      help="Configuration file.")
  parser.add_argument('-c', '--console', action='store_true',
      help="Enable console output.")
  parser.add_argument('-l', '--logfile', help="Enable output into logfile.")
  if args:
    parsed_argument = parser.parse_args(list(args))
  else:
    parsed_argument = parser.parse_args()
  logger_format = '%(asctime)s %(name)-13s: %(levelname)-8s %(message)s'
  formatter = logging.Formatter(logger_format)
  logging.basicConfig(level=logging.INFO,
                     format=logger_format)
  logger = logging.getLogger('erp5testnode')
  if parsed_argument.console or parsed_argument.logfile:
    if parsed_argument.console:
      logger.addHandler(logging.StreamHandler())
      logger.info('Activated console output.')
    if parsed_argument.logfile:
      file_handler = logging.handlers.RotatingFileHandler(
        filename=parsed_argument.logfile,
        maxBytes=20000000, backupCount=4)
      file_handler.setFormatter(formatter)
      logger.addHandler(file_handler)
      logger.info('Activated logfile %r output' % parsed_argument.logfile)
      CONFIG['log_file'] = parsed_argument.logfile
  else:
    logger.addHandler(logging.NullHandler())
  CONFIG['logger'] = logger.info
  config = ConfigParser.SafeConfigParser()
  # do not change case of option keys
  config.optionxform = str
  config.readfp(parsed_argument.configuration_file[0])
  for key in ('slapos_directory','working_directory','test_suite_directory',
              'log_directory','run_directory', 'srv_directory', 'proxy_host',
              'software_directory',
              'proxy_port', 'git_binary','zip_binary','node_quantity',
              'test_node_title', 'ipv4_address','ipv6_address','test_suite_master_url',
              'slapgrid_partition_binary','slapgrid_software_binary',
              'slapproxy_binary', 'httpd_ip', 'httpd_port', 'httpd_software_access_port',
              'computer_id', 'server_url'):
    CONFIG[key] = config.get('testnode',key)

  for key in ('slapos_directory', 'working_directory', 'test_suite_directory',
      'log_directory', 'run_directory', 'srv_directory', 'software_directory'):
    d = CONFIG[key]
    if not os.path.isdir(d):
      raise ValueError('Directory %r does not exists.' % d)
  CONFIG['master_url'] = 'http://%s:%s' % (CONFIG['proxy_host'],
        CONFIG['proxy_port'])
  CONFIG['httpd_url'] = 'https://[%s]:%s' % (CONFIG['httpd_ip'],
        CONFIG['httpd_port'])
  CONFIG['system_temp_folder'] = "/tmp"

  # generate vcs_repository_list
  if 'bot_environment' in config.sections():
    bot_environment = dict(config.items('bot_environment'))
  else:
    bot_environment = {}
  CONFIG['bot_environment'] = bot_environment
  CONFIG['environment'] = dict(config.items('environment'))
  if 'instance_dict' in config.sections():
    instance_dict = dict(config.items('instance_dict'))
  else:
    instance_dict = {}
  CONFIG['instance_dict'] = instance_dict
  if 'software_list' in config.sections():
    CONFIG['software_list'] = filter(None,
        config.get("software_list", "path_list").split(","))
  
  testnode = TestNode(logger.info, CONFIG)
  testnode.run()



##############################################################################
#
# Copyright (c) 2012 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
"""
Client implementation for portal_task_distribution.

Example use:
  import erp5.util.taskdistribution
  tool = erp5.util.taskdistribution.TaskDistributionTool(...)
  test_result = tool.createTestResult(...)
  test_result.addWatch('foo', open('foo'))
  while True:
      test_line = test_result.start()
      if not test_line:
          break
      # Run the test_line.name test
      test_line.stop()
"""
import httplib
import logging
import select
import socket
import threading
import time
import xmlrpclib

__all__ = ['TaskDistributionTool', 'TestResultProxy', 'TestResultLineProxy', 'patchRPCParser']

# Depending on used xmlrpc backend, different exceptions can be thrown.
SAFE_RPC_EXCEPTION_LIST = [socket.error, xmlrpclib.ProtocolError,
    xmlrpclib.Fault, httplib.BadStatusLine, httplib.ResponseNotReady]
parser, _ = xmlrpclib.getparser()
if xmlrpclib.ExpatParser and isinstance(parser, xmlrpclib.ExpatParser):
    SAFE_RPC_EXCEPTION_LIST.append(xmlrpclib.expat.ExpatError)
else:
    import sys
    print >> sys.stderr, 'Warning: unhandled xmlrpclib parser %r, some ' \
        'exceptions might get through safeRpcCall' % (parser, )
    del sys
SAFE_RPC_EXCEPTION_LIST = tuple(SAFE_RPC_EXCEPTION_LIST)
del parser, _

def null_callable(*args, **kw):
    pass

class NullLogger(object):
    def __getattr__(self, name):
        return null_callable
null_logger = NullLogger()

def patchRPCParser(error_handler):
    """
    Patch xmlrpcmlib's parser class, so it logs data content in case of errors,
    to ease debugging.
    Warning: this installs a monkey patch on a generic class, so it's last
    comes wins. Must *not* be enabled by default.

    error_handler (callable)
      Receives the erroneous data as first parameter, and the exception
      instance as second parameter.
      If it returns a false value (ie, handler did not recover from the error),
      exception is re-raised.
    """
    parser, _ = xmlrpclib.getparser()
    parser_klass = parser.__class__
    original_feed = parser_klass.feed
    def verbose_feed(self, data):
        try:
            return original_feed(self, data)
        except Exception, exc:
            if not error_handler(data, exc):
                raise
    parser_klass.feed = verbose_feed

class RPCRetry(object):
    def __init__(self, proxy, retry_time, logger):
        super(RPCRetry, self).__init__()
        self._proxy = proxy
        self._retry_time = retry_time
        self._logger = logger

    def _RPC(self, func_id, args=()):
        return getattr(self._proxy, func_id)(*args)

    def _retryRPC(self, func_id, args=()):
        retry_time = self._retry_time
        while True:
            try:
                return self._RPC(func_id, args)
            except SAFE_RPC_EXCEPTION_LIST:
                self._logger.warning('Got exception, retrying: %s%r '
                    'in %is', func_id, tuple(args), retry_time, exc_info=1)
                time.sleep(retry_time)
                retry_time *= 1.5

class TestResultLineProxy(RPCRetry):
    """
    Represents a single test in a suite.

    Properties:
    name (str) (ro)
      Test name, as provided to TaskDistributionTool.createTestResult .
    """
    def __init__(self, proxy, retry_time, logger, test_result_line_path,
            test_name):
        super(TestResultLineProxy, self).__init__(proxy, retry_time, logger)
        self._test_result_line_path = test_result_line_path
        self._name = test_name

    def __repr__(self):
        return '<%s(%r, %r) at %x>' % (self.__class__.__name__,
            self._test_result_line_path, self._name, id(self))

    @property
    def name(self):
        return self._name

    def isTestCaseAlive(self):
        """
        Tell if test result line is still alive on site.
        """
        try:
          return bool(self._retryRPC('isTestCaseAlive', [self._test_result_line_path]))
        except:
          raise ValueError('isTestCaseAlive Failed.')

    def stop(self, test_count=None, error_count=None, failure_count=None,
            skip_count=None, duration=None, date=None, command=None,
            stdout=None, stderr=None, html_test_result=None, **kw):
        """
        Notify server of test completion.

        Without any parameter, notifies of a test failure which prevents any
        precise reading (step count, how many succeeded, etc).

        BBB: extra named arguments are deprecated (if some are really needed,
        they must be declared as explicit parameters, with proper default
        value).
        """
        status_dict = dict(x for x in (
            ('test_count', test_count),
            ('error_count',  error_count),
            ('failure_count',  failure_count),
            ('skip_count',  skip_count),
            ('duration',  duration),
            ('date',  date),
            ('command',  command),
            ('stdout',  stdout),
            ('stderr',  stderr),
            ('html_test_result',  html_test_result),
        ) if x[1] is not None)
        if kw:
            self._logger.info('Extra parameters provided: %r', kw)
            status_dict.update(kw)
        self._retryRPC('stopUnitTest', (self._test_result_line_path,
            status_dict))

class TestResultProxy(RPCRetry):
    """
    Represents a test suite run.

    Allows fetching work to do (eg a single test in an entire run), monitoring
    log files, informing server of problems and monitoring server-side
    cancellation.

    Properties
    watcher_period (float) (rw)
      How long log watcher sleeps between successive uploading latest data
      chunks.
    revision (str) (ro)
      Revision to test. Might be different from the revision requested, when a
      test batch is running on an older revision.
    """
    _watcher_can_run = True
    _watcher_thread = None

    def __init__(self, proxy, retry_time, logger, test_result_path, node_title,
            revision):
        super(TestResultProxy, self).__init__(proxy, retry_time, logger)
        self._test_result_path = test_result_path
        self._node_title = node_title
        self._revision = revision
        self._watcher_period = 60
        self._watcher_dict = {}
        self._watcher_condition = threading.Condition()
    def __repr__(self):
        return '<%s(%r, %r, %r) at %x>' % (self.__class__.__name__,
            self._test_result_path, self._node_title, self._revision, id(self))

    @property
    def test_result_path(self):
        return self._test_result_path

    @property
    def revision(self):
        return self._revision

    def start(self, exclude_list=()):
        """
        Ask for a test to run, among the list of tests composing this test
        result.
        Return an TestResultLineProxy instance, or None if there is nothing to
        do.
        """
        result = self._retryRPC('startUnitTest', (self._test_result_path,
            exclude_list))
        if result:
            line_url, test_name = result
            result = TestResultLineProxy(self._proxy, self._retry_time,
                self._logger, line_url, test_name)
        return result

    def reportFailure(self, date=None, command=None, stdout=None, stderr=None):
        """
        Report a test-node-level problem, preventing the test from continuing
        on this node.
        """
        self._stopWatching()
        status_dict = {
            'date': date,
        }
        if command is not None:
            status_dict['command'] = command
        if stdout is not None:
            status_dict['stdout'] = stdout
        if stderr is not None:
            status_dict['stderr'] = stderr
        self._retryRPC('reportTaskFailure', args=(self._test_result_path,
            status_dict, self._node_title))

    def reportStatus(self, command, stdout, stderr):
        """
        Report some progress.

        Used internally by file monitoring, you shouldn't have to use this
        directly.
        """
        try:
            self._RPC('reportTaskStatus', (self._test_result_path, {
                'command': command,
                'stdout': stdout,
                'stderr': stderr,
            }, self._node_title))
        except SAFE_RPC_EXCEPTION_LIST:
            self._logger.warning('Got exception in reportTaskStatus, giving up',
                exc_info=1)

    def isAlive(self):
        """
        Tell if test is still alive on site.

        Useful to probe for test cancellation by user, so a new test run can
        be started without waiting for current one to finish.
        """
        try:
            return self._RPC('isTaskAlive', (self._test_result_path, ))
        except SAFE_RPC_EXCEPTION_LIST:
            self._logger.warning('Got exception in isTaskAlive, assuming alive',
                exc_info=1)
            return 1

    @property
    def watcher_period(self):
        return self._watcher_period

    @watcher_period.setter
    def watcher_period(self, period):
        cond = self._watcher_condition
        with cond:
            self._watcher_period = period
            cond.notify()

    def addWatch(self, name, stream, max_history_bytes=None):
        """
        Monitor given file, sending a few latest lines to remote server.
        name (any)
          Arbitrary identifier for stream. Must be usable as a dict key.
        stream (file object)
          File to monitor from its current offset.
        max_history_bytes (int, None)
          How many bytes to send to remote server at most for each wakeup.
          If None, send all lines.
        """
        watcher_dict = self._watcher_dict
        if not watcher_dict:
            self._startWatching()
        elif name in watcher_dict:
            raise ValueError('Name already known: %r' % (name, ))
        watcher_dict[name] = (stream, max_history_bytes)

    def removeWatch(self, name):
        """
        Stop monitoring given stream.
        """
        watcher_dict = self._watcher_dict
        del watcher_dict[name]
        if not watcher_dict:
            self._stopWatching()

    def _startWatching(self):
        if self._watcher_thread is not None:
            raise ValueError('Thread already started')
        self._watcher_thread = thread = threading.Thread(target=self._watcher)
        thread.daemon = True
        thread.start()

    def _watcher(self):
        cond = self._watcher_condition
        while self._watcher_can_run and self.isAlive():
            caption_list = []
            append = caption_list.append
            for name, (stream, max_history_bytes) in \
                    self._watcher_dict.iteritems():
                append('==> %s <==' % (name, ))
                start = stream.tell()
                stream.seek(0, 2)
                end = stream.tell()
                if start == end:
                    caption = time.strftime(
                        '(no new lines at %Y/%m/%d %H:%M:%S)', time.gmtime())
                else:
                    to_read = end - start
                    if to_read < 0:
                        # File got truncated, treat the whole content as new.
                        to_read = end
                    if max_history_bytes is not None:
                        to_read = min(to_read, max_history_bytes)
                    stream.seek(-to_read, 1)
                    caption = stream.read(to_read)
                append(caption)
            self.reportStatus('', '\n'.join(caption_list), '')
            with cond:
                cond.wait(self._watcher_period)

    def _stopWatching(self):
        cond = self._watcher_condition
        with cond:
            self._watcher_can_run = False
            cond.notify()
        if self._watcher_thread is not None:
          self._watcher_thread.join()

    def stop(self):
        """
        
        """
        return self._retryRPC('stopTest', [self._test_result_path])
        
    def getRunningTestCase(self):
      """
      Return the relative path of the next test with the running state
      """
      return self._retryRPC('getRunningTestCase', [self._test_result_path])

class ServerProxy(xmlrpclib.ServerProxy):

    def __init__(self, *args, **kw):
        xmlrpclib.ServerProxy.__init__(self, *args, **kw)
        transport = self.__transport
        def make_connection(*args, **kw):
            conn = transport.__class__.make_connection(transport, *args, **kw)
            # BBB: On Python < 2.7, HTTP connection is wrapped
            c = getattr(conn, '_conn', conn)
            assert hasattr(c, 'timeout')
            c.timeout = 120
            return conn
        transport.make_connection = make_connection
        self.__rpc_lock = threading.Lock()

    def __request(self, *args, **kw):
        with self.__rpc_lock:
            return xmlrpclib.ServerProxy.__request(self, *args, **kw)

class TaskDistributionTool(RPCRetry):
    def __init__(self, portal_url, retry_time=64, logger=None):
        """
        portal_url (str, None)
          Portal URL of ERP5 site to use as a task distributor.
          If None, single node setup is assumed.
        """
        if logger is None:
            logger = null_logger
        if portal_url is None:
            proxy = DummyTaskDistributionTool()
        else:
            proxy = ServerProxy(
                portal_url,
                allow_none=True,
            ).portal_task_distribution
        super(TaskDistributionTool, self).__init__(proxy, retry_time, logger)
        protocol_revision = self._retryRPC('getProtocolRevision')
        if protocol_revision != 1:
            raise ValueError('Unsupported protocol revision: %r',
                protocol_revision)

    def createTestResult(self, revision, test_name_list, node_title,
            allow_restart=False, test_title=None, project_title=None):
        """
        (maybe) create a new test run.
        revision (str)
          An opaque string describing code being tested.
        test_name_list (list of str)
          List of tests being part of this test run. May be empty.
        node_title (str)
          Human-readable test node identifier, so an adnmin can know which
          node does what.
        allow_restart (bool)
          When true, a tet result is always created, even if a former finished
          one is found for same name and revision pair.
        test_title (str)
          Human-readable title for test. Must be identical for successive runs.
          Allows browsing its result history.
        project_title (str)
          Existing project title, so test result gets associated to it.

        Returns None if no test run is needed (a test run for given name and
        revision has already been completed).
        Otherwise, returns a TestResultProxy instance.
        """
        result = self._retryRPC('createTestResult', ('', revision,
            test_name_list, allow_restart, test_title, node_title,
            project_title))
        if result:
            test_result_path, revision = result
            result = TestResultProxy(self._proxy, self._retry_time,
                self._logger, test_result_path, node_title, revision)
        return result

class TaskDistributor(RPCRetry):

    def __init__(self,portal_url,retry_time=64,logger=None):
        if logger is None:
           logger = null_logger
        if portal_url is None:
            proxy = DummyTaskDistributionTool()
        else:
            proxy = ServerProxy(portal_url, allow_none=True)
        super(TaskDistributor, self).__init__(proxy, retry_time,logger)
        protocol_revision = self._retryRPC('getProtocolRevision')
        if protocol_revision != 1:
            raise ValueError('Unsupported protocol revision: %r',
                protocol_revision)

    def startTestSuite(self,node_title,computer_guid='unknown'):
      """
        Returns None if no test suite is needed.
        therwise, returns a JSON with all the test suite parameters.
      """
      result = self._retryRPC('startTestSuite',(node_title,computer_guid,))
      return result

    def getTestType(self):
      """
        Return the Test Type
      """
      result = self._retryRPC('getTestType')
      return result

    def subscribeNode(self, node_title, computer_guid):
      """
        Susbscribes node with the node title and the computer guid.
      """
      self._retryRPC('subscribeNode', (node_title,computer_guid,))


    def generateConfiguration(self, test_suite_title):
      """
        Generates a configuration from a test_suite_title
      """
      return self._retryRPC('generateConfiguration', (test_suite_title,))


    def isMasterTestnode(self, test_node_title):
      """
        Returns True or False if the testnode is the master
      """
      return self._retryRPC('isMasterTestnode', (test_node_title,))

    def getSlaposAccountKey(self):
      """
        Returns the slapos account key related to the distributor
      """
      return self._retryRPC('getSlaposAccountKey')
    
    def getSlaposAccountCertificate(self):
      """
        Returns the slapos account certificate related to the distributor
      """
      return self._retryRPC('getSlaposAccountCertificate')

    def getSlaposUrl(self):
      """
        Returns the url of slapos master related to the distributor
      """
      return self._retryRPC('getSlaposUrl')
      
    def getSlaposHateoasUrl(self):
      """
        Returns the url of API REST using hateoas of
        slapos master related to the distributor
      """
      return self._retryRPC('getSlaposHateoasUrl')


class DummyTaskDistributionTool(object):
    """
    Fake remote server.

    Useful when willing to locally run all tests without reporting to any
    server.

    This class should remain internal to this module.
    """
    test_name_list = None

    def __init__(self):
        self._lock = threading.Lock()

    def getProtocolRevision(self):
        return 1

    def createTestResult(self, name, revision, test_name_list, *args):
        self.test_name_list = test_name_list[:]
        return None, revision

    def startUnitTest(self, test_result_path, exclude_list=()):
        with self._lock:
            for i, test in enumerate(self.test_name_list):
                if test not in exclude_list:
                    del self.test_name_list[i]
                    return None, test

    def stopUnitTest(self, *args):
        pass

    reportTaskFailure = reportTaskStatus = stopUnitTest

    def isTaskAlive(self, *args):
        return int(bool(self.test_name_list))




import re, sys, threading, os, subprocess
import traceback
import errno
from pprint import pprint

_format_command_search = re.compile("[[\\s $({?*\\`#~';<>&|]").search
_format_command_escape = lambda s: "'%s'" % r"'\''".join(s.split("'"))
def format_command(*args, **kw):
  cmdline = []
  for k, v in sorted(kw.items()):
    if _format_command_search(v):
      v = _format_command_escape(v)
    cmdline.append('%s=%s' % (k, v))
  for v in args:
    if _format_command_search(v):
      v = _format_command_escape(v)
    cmdline.append(v)
  return ' '.join(cmdline)

def subprocess_capture(p, quiet=False):
  def readerthread(input, output, buffer):
    while True:
      data = input.readline()
      if not data:
        break
      output(data)
      buffer.append(data)
  if p.stdout:
    stdout = []
    output = quiet and (lambda data: None) or sys.stdout.write
    stdout_thread = threading.Thread(target=readerthread,
                                     args=(p.stdout, output, stdout))
    stdout_thread.setDaemon(True)
    stdout_thread.start()
  if p.stderr:
    stderr = []
    stderr_thread = threading.Thread(target=readerthread,
                                     args=(p.stderr, sys.stderr.write, stderr))
    stderr_thread.setDaemon(True)
    stderr_thread.start()
  if p.stdout:
    stdout_thread.join()
  if p.stderr:
    stderr_thread.join()
  p.wait()
  return (p.stdout and ''.join(stdout),
          p.stderr and ''.join(stderr))

class SubprocessError(EnvironmentError):
  def __init__(self, status_dict):
    self.status_dict = status_dict
  def __getattr__(self, name):
    return self.status_dict[name]
  def __str__(self):
    return 'Error %i' % self.status_code

class Persistent(object):
  """Very simple persistent data storage for optimization purpose

  This tool should become a standalone daemon communicating only with an ERP5
  instance. But for the moment, it only execute 1 test suite and exists,
  and test suite classes may want some information from previous runs.
  """

  def __init__(self, filename):
    self._filename = filename

  def __getattr__(self, attr):
    if attr == '_db':
      try:
        db = file(self._filename, 'r+')
      except IOError, e:
        if e.errno != errno.ENOENT:
          raise
        db = file(self._filename, 'w+')
      else:
        try:
          self.__dict__.update(eval(db.read()))
        except StandardError:
          pass
      self._db = db
      return db
    self._db
    return super(Persistent, self).__getattribute__(attr)

  def sync(self):
    self._db.seek(0)
    db = dict(x for x in self.__dict__.iteritems() if x[0][:1] != '_')
    pprint.pprint(db, self._db)
    self._db.truncate()

class TestSuite(object):
  """
  Subclasses may redefine the following properties:
  mysql_db_count (integer, >=1)
    Maximum number of SQL databases connection strings needed by any tests ran
    in this suite. Tests will get mysql_db_count - 1 connection strings in
    extra_sql_connection_string_list environment variable.
  """

  RUN_RE = re.compile(
    r'Ran (?P<all_tests>\d+) tests? in (?P<seconds>\d+\.\d+)s',
    re.DOTALL)

  STATUS_RE = re.compile(r"""
    (OK|FAILED)\s+\(
      (failures=(?P<failures>\d+),?\s*)?
      (errors=(?P<errors>\d+),?\s*)?
      (skipped=(?P<skips>\d+),?\s*)?
      (expected\s+failures=(?P<expected_failures>\d+),?\s*)?
      (unexpected\s+successes=(?P<unexpected_successes>\d+),?\s*)?
    \)
    """, re.DOTALL | re.VERBOSE)

  SUB_STATUS_RE = re.compile(
      r"""SUB\s+RESULT:\s+(?P<all_tests>\d+)\s+Tests,\s+
      (?P<failures>\d+)\s+Failures\s*
      \(?
        (skipped=(?P<skips>\d+),?\s*)?
        (expected\s+failures=(?P<expected_failures>\d+),?\s*)?
        (unexpected\s+successes=(?P<unexpected_successes>\d+),?\s*)?
      \)?
      """,
      re.DOTALL | re.VERBOSE)

  mysql_db_count = 1
  allow_restart = False
  realtime_output = True
  stdin = file(os.devnull)

  def __init__(self, max_instance_count, **kw):
    self.__dict__.update(kw)
    self._path_list = ['tests']
    pool = threading.Semaphore(max_instance_count)
    self.acquire = pool.acquire
    self.release = pool.release
    self._instance = threading.local()
    self._pool = max_instance_count == 1 and [None] or \
                 range(1, max_instance_count + 1)
    self._ready = set()
    self.running = {}
    if max_instance_count != 1:
      self.realtime_output = False
    elif os.isatty(1):
      self.realtime_output = True
    self.persistent = Persistent('run_test_suite-%s.tmp'
                                 % self.__class__.__name__)

  instance = property(lambda self: self._instance.id)

  def start(self, test, on_stop=None):
    assert test not in self.running
    self.running[test] = instance = self._pool.pop(0)
    def run():
      try:
        self._instance.id = instance
        if instance not in self._ready:
          self._ready.add(instance)
          self.setup()
        status_dict = self.run(test)
        if on_stop is not None:
          on_stop(status_dict)
        self._pool.append(self.running.pop(test))
      finally:
        self.release()
    thread = threading.Thread(target=run)
    thread.setDaemon(True)
    thread.start()

  def update(self):
    self.checkout() # by default, update everything

  def setup(self):
    pass

  def run(self, test):
    raise NotImplementedError

  def getTestList(self):
    raise NotImplementedError

  def spawn(self, *args, **kw):
    quiet = kw.pop('quiet', False)
    env = kw and dict(os.environ, **kw) or None
    command = format_command(*args, **kw)
    print '\n$ ' + command
    sys.stdout.flush()
    try:
      p = subprocess.Popen(args, stdin=self.stdin, stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE, env=env)
    except Exception:
      # Catch any exception here, to warn user instead of beeing silent,
      # by generating fake error result
      result = dict(status_code=-1,
                    command=command,
                    stderr=traceback.format_exc(),
                    stdout='')
      raise SubprocessError(result)
    if self.realtime_output:
      stdout, stderr = subprocess_capture(p, quiet)
    else:
      stdout, stderr = p.communicate()
      if not quiet:
        sys.stdout.write(stdout)
      sys.stderr.write(stderr)
    result = dict(status_code=p.returncode, command=command,
                  stdout=stdout, stderr=stderr)
    if p.returncode:
      raise SubprocessError(result)
    return result

# (XXX) The code bellow is an generic extension to run a test for any egg. 
#       The code above was moved from ERP5 code base, because it is generic
#       Enough to be used by others.
# (FIXME) Imports should be reorganised in a better way
import argparse
import sys
import glob
import re
import os
import shlex
from erp5.util import taskdistribution
import re, os, shlex, glob

class EggTestSuite(TestSuite):

  def run(self, test):
    print test
    original_dir = os.getcwd()
    try:
      os.chdir(test)
      return self.runUnitTest(test)
    finally:
      os.chdir(original_dir)

  def runUnitTest(self, *args, **kw):
    try:
      # (FIXME) The python should be provided by environment with 
      #         appropriated configuration.
      runUnitTest = "python setup.py test"
      args = tuple(shlex.split(runUnitTest))
      status_dict = self.spawn(*args, **kw)
    except SubprocessError, e:
      status_dict = e.status_dict
    test_log = status_dict['stderr']
    search = self.RUN_RE.search(test_log)
    if search:
      groupdict = search.groupdict()
      status_dict.update(duration=float(groupdict['seconds']),
                         test_count=int(groupdict['all_tests']))
    search = self.STATUS_RE.search(test_log)
    def updateStatusDictWithGroup(groupdict):
      status_dict.update(error_count=int(groupdict.get('errors') or 0),
                         failure_count=int(groupdict['failures'] or 0)
                                 +int(groupdict['unexpected_successes'] or 0),
                         skip_count=int(groupdict['skips'] or 0)
                                   +int(groupdict['expected_failures'] or 0))
    if search:
      groupdict = search.groupdict()
      updateStatusDictWithGroup(groupdict)
    # In case the test is used to launch another one (like functional test),
    # look for results inside stderr
    sub_search = self.SUB_STATUS_RE.search(status_dict['stdout'])
    if sub_search:
      groupdict = sub_search.groupdict()
      updateStatusDictWithGroup(groupdict)
      status_dict.update(test_count=int(groupdict['all_tests']))
    return status_dict

  def getTestList(self):
    # (FIXME) The test name should be nicer in order to provide a good report.
    #         On task distribution.
    source_code_to_test = os.environ.get("SOURCE_CODE_TO_TEST", '.')
    return source_code_to_test.split(",")

def runTestSuite():
  parser = argparse.ArgumentParser(description='Run a test suite.')
  parser.add_argument('--test_suite', help='The test suite name')
  parser.add_argument('--test_suite_title', help='The test suite title',
                      default=None)
  parser.add_argument('--test_node_title', help='The test node title',
                      default=None)
  parser.add_argument('--project_title', help='The project title',
                      default=None)
  parser.add_argument('--revision', help='The revision to test',
                      default='dummy_revision')
  parser.add_argument('--node_quantity', help='Number of parallel tests to run',
                      default=1, type=int)
  parser.add_argument('--master_url',
                      help='The Url of Master controling many suites',
                      default=None)
  parser.add_argument('--source_code_path_list',
                      help='List of Eggs folders to test, splited by commam',
                      default='.')

  args = parser.parse_args()
  master = taskdistribution.TaskDistributionTool(args.master_url)
  os.environ.setdefault("SOURCE_CODE_TO_TEST", args.source_code_path_list)
  test_suite_title = args.test_suite_title or args.test_suite
  revision = args.revision
  suite = EggTestSuite(1, test_suite=args.test_suite,
                    node_quantity=args.node_quantity,
                    revision=revision)

  test_result = master.createTestResult(revision, suite.getTestList(),
    args.test_node_title, suite.allow_restart, test_suite_title,
    args.project_title)
  if test_result is not None:
    assert revision == test_result.revision, (revision, test_result.revision)
    while suite.acquire():
      test = test_result.start(suite.running.keys())
      if test is not None:
        suite.start(test.name, lambda status_dict, __test=test:
          __test.stop(**status_dict))
      elif not suite.running:
        break

